@@ -1,9 +1,66 @@
-import logging
-
 import tensorflow as tf
 from tensorflow.keras import layers
 
 
+class RBFExpansion(layers.Layer):
+    def __init__(self,
+                 dimension=128,
+                 init_gap=10,
+                 init_max_distance=7,
+                 trainable=False):
+        """ Layer to calculate radial basis function 'embeddings' for a continuous input variable. The width and
+        location of each bin can be optionally trained. Essentially equivalent to a 1-hot embedding for a continous
+        variable.
+
+        Parameters
+        ----------
+        dimension: The total number of distance bins
+        init_gap: The initial width of each gaussian distribution
+        init_max_distance: the initial maximum value of the continuous variable
+        trainable: Whether the centers and gap parameters should be added as trainable NN parameters.
+        """
+        super(RBFExpansion, self).__init__()
+        self.init_gap = init_gap
+        self.init_max_distance = init_max_distance
+        self.dimension = dimension
+        self.trainable = trainable
+
+    def build(self, input_shape):
+        self.centers = tf.Variable(
+            name='centers',
+            initial_value=tf.range(0,
+                                   self.init_max_distance,
+                                   delta=self.init_max_distance /
+                                   self.dimension),
+            trainable=self.trainable,
+            dtype=tf.float32)
+
+        self.gap = tf.Variable(name='gap',
+                               initial_value=tf.constant(self.init_gap,
+                                                         dtype=tf.float32),
+                               trainable=self.trainable,
+                               dtype=tf.float32)
+
+    def call(self, inputs, **kwargs):
+        distances = tf.where(tf.math.is_nan(inputs),
+                             tf.zeros_like(inputs, dtype=inputs.dtype), inputs)
+        offset = tf.expand_dims(distances, -1) - tf.cast(
+            self.centers, inputs.dtype)
+        logits = -self.gap * offset**2
+        return tf.exp(logits)
+
+    def compute_mask(self, inputs, mask=None):
+        return tf.logical_not(tf.math.is_nan(inputs))
+
+    def get_config(self):
+        return {
+            'init_gap': self.init_gap,
+            'init_max_distance': self.init_max_distance,
+            'dimension': self.dimension,
+            'trainable': self.trainable
+        }
+
+
 def batched_segment_op(data,
                        segment_ids,
                        num_segments,
@@ -42,22 +99,22 @@ def batched_segment_op(data,
     return tf.reshape(reduced_data, [batch_size, num_segments, data.shape[-1]])
 
 
-class Slice(layers.Layer):
-    def __init__(self, slice_obj, *args, **kwargs):
-        super(Slice, self).__init__(*args, **kwargs)
-        self.slice_obj = slice_obj
-        self.supports_masking = True
-
-    def call(self, inputs, mask=None):
-        return inputs[self.slice_obj]
-
-    def get_config(self):
-        return {'slice_obj': str(self.slice_obj)}
-
-    @classmethod
-    def from_config(cls, config):
-        config['slice_obj'] = eval(config['slice_obj'])
-        return cls(**config)
+# class Slice(layers.Layer):
+#     def __init__(self, slice_obj, *args, **kwargs):
+#         super(Slice, self).__init__(*args, **kwargs)
+#         self.slice_obj = slice_obj
+#         self.supports_masking = True
+#
+#     def call(self, inputs, mask=None):
+#         return inputs[self.slice_obj]
+#
+#     def get_config(self):
+#         return {'slice_obj': str(self.slice_obj)}
+#
+#     @classmethod
+#     def from_config(cls, config):
+#         config['slice_obj'] = eval(config['slice_obj'])
+#         return cls(**config)
 
 
 class Gather(layers.Layer):
@@ -82,16 +139,19 @@ class Reduce(layers.Layer):
 
         return data, segment_ids, target, data_mask
 
+    def compute_output_shape(self, input_shape):
+        data_shape, _, target_shape = input_shape
+        return [data_shape[0], target_shape[1], data_shape[-1]]
+
     def call(self, inputs, mask=None):
         data, segment_ids, target, data_mask = self._parse_inputs_and_mask(
             inputs, mask)
         num_segments = tf.shape(target, out_type=segment_ids.dtype)[1]
-        return batched_segment_op(
-            data,
-            segment_ids,
-            num_segments,
-            data_mask=data_mask,
-            reduction=self.reduction)
+        return batched_segment_op(data,
+                                  segment_ids,
+                                  num_segments,
+                                  data_mask=data_mask,
+                                  reduction=self.reduction)
 
     def get_config(self):
         return {'reduction': self.reduction}
@@ -100,7 +160,6 @@ class Reduce(layers.Layer):
 class ConcatDense(layers.Layer):
     """ Layer to combine the concatenation and two dense layers. Just useful as a common operation in the graph
     layers """
-
     def __init__(self, **kwargs):
         super(ConcatDense, self).__init__(**kwargs)
         self.supports_masking = True
